/*
 * Copyright (c) 2002 br1an <brian@innu.org>
 * All rights reserved.
 *
 * Official Stephanie homepage:
 *
 *	http://innu.org/~brian/Stephanie
 *
 * Shoutz go to shegget, S`, _eci, doe, jaf.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/kern_acl.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <pwd.h>
#include <sysexits.h>
#include <syslog.h>

void usage(void) {
	extern char *__progname;

	(void) fprintf(stderr, "Stephanie: ACL: Trusted users list "
	    "administration utility\n");

	(void) fprintf(stderr, "usage: %s <options>\n"
	    "  -a user    Add a user to the trust list\n"
	    "  -d user    Delete a user from the list\n"
	    "  -s         Show the current trust list\n"
	    "(``user'' can be either a username or a UID)\n", __progname);
}

int inputtype(char *p) {
	if (isdigit(*(p++))) {
		while ((*p) && isdigit(*p)) p++;
		if (!(*p)) return (0);
	}

	return (-1);
}

int main(int argc, char **argv) {
	uid_t uid, list[ACL_SIZE];
	int c, i;
	extern char *__progname;
	struct passwd *pwd;

	if (geteuid() && getuid())
		errx(EX_NOPERM, "Superuser access required.");

	if (argc == 1 || argc > 3) {
		usage();
		exit(EX_USAGE);
	}

	openlog(__progname, LOG_PID, LOG_AUTH);

	while ((c = getopt(argc, argv, "a:d:l:s")) != -1) {
		switch (c) {
		case 'a':
		case 'd':
			if (!inputtype(optarg)) {
				uid = (uid_t) atol(optarg);

				pwd = getpwuid(uid);
				if (!pwd && (c == 'a'))
					errx(EX_NOUSER, "UID %u does not"
					    " exist.", uid);
			} else {
				pwd = getpwnam(optarg);
				if (!pwd)
					errx(EX_NOUSER, "User \"%s\" does"
					    " not exist.", optarg);

				uid = pwd->pw_uid;
			}

			if (uid == 0)
				errx(EX_DATAERR, "Superuser %s", 
				    (c == 'a') ? "added by default." : 
				    "may not be removed.");

			if (c == 'a') {
				if (syscall(SYS_aclctl, ACL_ADD, 
				    uid, NULL) == -1) {
					syslog(LOG_NOTICE, "Full trust "
					    "list.\n");
					errx(EX_DATAERR, "Full trust "
					    "list.");
				}

				syslog(LOG_NOTICE, "UID %u (%s) added to "
				    "trust list.\n", uid, (pwd) ? 
				    pwd->pw_name : "no username");
				(void) printf("%s: UID %u (%s) added to "
				    "trust list.\n", __progname, uid, 
				    (pwd) ? pwd->pw_name : "no username");
			} else {
				if (syscall(SYS_aclctl, ACL_REMOVE, uid, 
				    NULL) == -1) {
					errx(EX_NOUSER, "UID %u not found"
					    " in trust list.", uid);
				}

				syslog(LOG_NOTICE, "UID %u (%s) removed "
				    "from trust list.", uid, (pwd) ? 
				    pwd->pw_name : "no username");
				(void) printf("aclctl: UID %u (%s) " 
				    "removed from trust list.\n", uid, (pwd) 
				    ? pwd->pw_name : "no username");
			}

			break;

		case 's':
			/*
			 * It is Very Important that `list` is an array 
			 * of size ACL_SIZE -- the kernel expects this.
			 * Failure to do so can result in a panic. 
			 * However, only root can issue the aclctl system 
			 * call.
	                 */
			if (syscall(SYS_aclctl, ACL_SHOW, -1, list) == -1)
				err(EX_OSERR, "Hideous internal error.");

			(void) printf("Trusted users: ");
			for (i = 0; list[i] != ACL_INITIALIZER; i++) {
				pwd = getpwuid(list[i]);
				if (pwd) (void) printf("%s ", pwd->pw_name);
				else (void) printf("[%u] ", list[i]);
			}
			(void) putchar('\n');

			break;

		default:
			usage();
			exit(EX_USAGE);
		}
	}

	closelog();

	return (0);
}
