/*
 * Copyright (c) 2002 br1an <brian@innu.org>
 * All rights reserved.
 *
 * Official Stephanie homepage:
 *
 *	http://innu.org/~brian/Stephanie
 *
 * Shoutz go to shegget, S`, _eci, doe, jaf.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/uio.h>
#include <sys/kern_k5.h>
#include <sys/unistd.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <md5.h>
#include <sysexits.h>
#include <unistd.h>

void usage(void) {
	extern char *__progname;

	fprintf(stderr, "Stephanie: k5: Binary integrity verification "
			"userland management utility\n"
			"usage: %s <cmd> [program]\n"
			"  add		Add <program> to the database\n"
			"  del		Delete <program> from the database\n"
			"  info		Show information about <program>\n"
			"  list		List database entries\n", __progname);
}

int parse_cmd(char *cmd) {
	size_t len = strlen(cmd);

	if ((len < 3) || (len > 4)) return (-1);
	if (!strcmp(cmd, "add")) return (K5_ADD);
	if (!strcmp(cmd, "del")) return (K5_DEL);
	if (!strcmp(cmd, "info")) return (K5_GETINFO);
	if (!strcmp(cmd, "list")) return (K5_GETOFFSET);

	return (-1);
}

void entry_print(struct k5_info *ptr) {
	(void) printf("Path: %s (%u, %u)\nMD5 Checksum: %s\nMatches: "
	    "%qu\nMismatches: %qu\n", ptr->k5_path, ptr->k5_inode,
		ptr->k5_device, 
	    ptr->k5_md5, ptr->k5_matches, ptr->k5_mismatches);
}

int main(int argc, char **argv) {
	int cmd;
	struct stat sb;
	char path[FILENAME_MAX], md5[33];
	ino_t inode = 0;
	dev_t device = 0;
	struct k5_info *ptr;
	off_t offset;

	if (argc < 2) {
		usage();
		exit(EX_USAGE);
	}

	cmd = parse_cmd(argv[1]);
	if (cmd == -1) {
		usage();
		exit(EX_USAGE);
	}

	/*
	 * Prepare data to be passed to the kernel according to command.
	 */
	switch (cmd) {
		case K5_ADD:
			if (argc != 3) {
				usage();
				exit(EX_USAGE);
			}

			if (stat(argv[2], &sb) == -1) {
				perror("stat");
				exit(EX_USAGE);
			}
			inode = sb.st_ino;
			device = sb.st_dev;
			
			MD5File(argv[2], md5);

			if (!realpath(argv[2], path)) {
				perror("realpath");
				exit(EX_DATAERR);
			}

			syscall(SYS_k5ctl, cmd, inode, device, path, md5, 
			    NULL, 0);

			break;

		case K5_DEL:
			if (argc != 3) {
				usage();
				exit(EX_USAGE);
			}

			if (stat(argv[2], &sb) == -1) {
				perror("stat");
				exit(EX_USAGE);
			}
			inode = sb.st_ino;
			device = sb.st_dev;

			syscall(SYS_k5ctl, cmd, inode, device, NULL, NULL, NULL,
			    0);

			break;

		case K5_GETINFO:
			if (argc != 3) {
				usage();
				exit(EX_USAGE);
			}

			if (stat(argv[2], &sb) == -1) {
				perror("stat");
				exit(EX_USAGE);
			}
			inode = sb.st_ino;
			device = sb.st_dev;

			ptr = (struct k5_info *) 
			    malloc(sizeof(struct k5_info));
			if (!ptr) {
				perror("malloc");
				exit(EX_UNAVAILABLE);
			}
			memset(ptr, 0, sizeof(struct k5_info));
		
			if (!syscall(SYS_k5ctl, cmd, inode, device, NULL, md5,
			    ptr, 0))
				entry_print(ptr);

			break;

		case K5_GETOFFSET:
			if (argc != 2) {
				usage();
				exit(EX_USAGE);
			}

			ptr = (struct k5_info *) 
			    malloc(sizeof(struct k5_info));
			if (!ptr) {
				perror("malloc");
				exit(0);
			}
			memset(ptr, 0, sizeof(struct k5_info));

			offset = 0;
			while (syscall(SYS_k5ctl, cmd, 0, 0, NULL, NULL, 
			    ptr, offset) != -1) {
				entry_print(ptr);
				putchar('\n');
				offset++;
			}

			printf("Total entries in database: %qu\n", 
			    offset);

			break;
	}

	return (0);
}
