/*
 * Copyright (c) 2002 br1an <brian@innu.org>
 * All rights reserved.
 *
 * Official Stephanie homepage:
 *
 *	http://innu.org/~brian/Stephanie
 *
 * Shoutz go to shegget, S`, _eci, doe, jaf.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/kern_k5.h>
#include <sys/vnode.h>

/*
 * Wrapper around the initialization functions.
 */
void k5_init(void) {
	LIST_INIT(&k5_head);
	k5_size = 0;
	printf("Stephanie: k5: Binary integrity verification "
	    "initialized.\n");
}

/*
 * Calculate the MD5 digest for a given file, generate an ASCII string 
 * containing the MD5 checksum as if it was created using md5(1), and save 
 * it in the supplied pointer.
 */
int k5_md5(struct exec_package *epp, char *checksum) {
	int error = 0, i;
	off_t offset;
	size_t len, resid, update_size;
	char *buffer, digest[16];
	MD5_CTX ctx;
	struct proc *p = curproc;

	/*
	 * Set 'len' to the default block size. If the user
	 * wants to save up kernel memory (K5_SAVEKMEM is defined in the 
	 * kernel config), use half the size.
	 */
	len = epp->ep_vap->va_blocksize;
#ifdef K5_SAVEKMEM
	len /= 2;
#endif

	/*
	 * Initialize the MD5 context and allocate space for the
	 * file data buffer.
	 */
	MD5Init(&ctx);
	MALLOC(buffer, char *, len, M_EXEC, M_WAITOK);

	offset = 0;
	while (offset < epp->ep_vap->va_size) {
		/*
		 * Clear the buffer to read data into, and fill it.
		 */
		bzero(buffer, len);
		error = vn_rdwr(UIO_READ, epp->ep_vp, buffer, len, offset, 
		    UIO_SYSSPACE, IO_NODELOCKED, p->p_ucred, &resid, p);
		if (error) {
#ifdef K5_VERBOSE
			uprintf("Stephanie: k5: Error (%d) during read "
			    "from binary: Aborting.\n");
#endif
			goto k5_error;
		}

		/*
		 * We use 'len - resid' when we update the MD5 context
		 * so that if we read less than 'len' we don't update
		 * with NULL bytes.
		 */
		update_size = len - resid;
		MD5Update(&ctx, buffer, update_size);
		if (update_size < len) len = update_size;
		offset += len;
	}

	/*
	 * Clear the MD5 digest buffer and finalize the MD5 context.
	 */
	bzero(digest, sizeof(digest));
	MD5Final(digest, &ctx);

	/*
	 * Assume we get an allocated buffer.
	 */
	bzero(checksum, MD5_SUM_LEN);
	for (i = 0; i < 16; i++)
		(void) snprintf(checksum, MD5_SUM_LEN, "%s%02x", checksum,
		    digest[i] & ~0xffffff00);

k5_error:
	/*
	 * Free the file data buffer.
	 */
	FREE(buffer, M_EXEC);

	return (error);
}

/*
 * Traverse the linked list, searching for an entry with a path matching
 * the one in the supplied pointer 'path'. Return 1 if found, otherwise 0.
 */
struct k5_info *k5_search(ino_t inode, dev_t device) {
	struct k5_info *ptr;

	if (k5_size == 0) return (NULL);

	LIST_FOREACH(ptr, &k5_head, k5_list)
		if (ptr->k5_inode == inode && ptr->k5_device == device)
 			return (ptr);

	return (NULL);
}

/*
 * Receives exec_package and nameidata pointers and wraps around all the
 * functions performing the database search, etc. Returns 1 if 
 * verification succeeded, 0 otherwise. FIX ERROR RETURNS.
 */
int k5_verify(struct exec_package *epp) {
	int error;
	char runtime_sum[MD5_SUM_LEN];
	struct k5_info *ptr;
	struct stat sb;

	/*
	 * Get the inode for the file.
	 */
	error = vn_stat(epp->ep_ndp->ni_vp, &sb, curproc);
	if (error) {
#ifdef K5_VERBOSE
		uprintf("Stephanie: k5: Error (%d) doing vn_stat(): "
		    "Aborting.\n");
#endif
		return (0);
	}

	/*
	 * See if we have such an entry in our database.
	 */
	ptr = k5_search(sb.st_ino, sb.st_dev);
	if (ptr) {
		/*
		 * Calculate the MD5 checksum for the binary.
		 */
		if ((error = k5_md5(epp, runtime_sum))) {
			uprintf("Stephanie: k5: Error during checksum "
			    "calculation.\n");
			return (0);
		}

		/*
		 * Compare the checksums. On match, raise the counter
		 * and return a true value. Mismatch? Raise the counter
		 * and return false.
		 */
		if (!strcmp(ptr->k5_md5, runtime_sum)) {
			(ptr->k5_matches)++;
			return (1);
		} else {
			(ptr->k5_mismatches)++;
			return (0);
		}
	}

#ifdef K5_STRICT
	return (0);
#endif

	return (1);
}

/*
 * Allocate space for a new entry in the database, store the data inside 
 * it, reset the match counters, and return 0 on success.
 */
int k5_add(ino_t inode, dev_t device, char *path, char *md5) {
	struct k5_info *ptr;

	/*
	 * Allocate memory for the new entry and zero it.
	 */
	MALLOC(ptr, struct k5_info *, sizeof(struct k5_info), M_TEMP, M_NOWAIT);
	if (!ptr) goto k5_nomem;
	bzero(ptr, sizeof(struct k5_info));

	ptr->k5_inode = inode;
	ptr->k5_device = device;
	strlcpy(ptr->k5_path, path, MAXPATHLEN);
	strlcpy(ptr->k5_md5, md5, MD5_SUM_LEN);
	ptr->k5_matches = 0;
	ptr->k5_mismatches = 0;

	/*
	 * Kinda like a stack that grows up, LIFO structure..
	 */
	LIST_INSERT_HEAD(&k5_head, ptr, k5_list);

	if (k5_size == 0) k5_size = 1;
	else k5_size++;

	return (0);

k5_nomem:
	return (ENOMEM);
}

/*
 * Remove the specified entry from the database. This function can't fail
 * because otherwise we would'nt be entering it.
 */
void k5_del(struct k5_info *ptr) {
	if (k5_size == 0) return;
	LIST_REMOVE(ptr, k5_list);
	k5_size--;
}
