/*
 * Copyright (c) 2002 br1an <brian@innu.org>
 * All rights reserved.
 *
 * Official Stephanie homepage:
 *
 *	http://innu.org/~brian/Stephanie
 *
 * Shoutz go to shegget, S`, _eci, doe, jaf.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <sys/kern_k5.h>
#include <sys/mount.h>
#include <sys/syscallargs.h>

int sys_k5ctl(p, v, retval)
	struct proc *p;
	void *v;
	register_t *retval;
{
	register struct sys_k5ctl_args /* {
		syscallarg(int) cmd;
		syscallarg(ino_t) inode;
		syscallarg(dev_t) device;
		syscallarg(char *) path;
		syscallarg(char *) md5;
		syscallarg(struct k5_info *) ptr;
		syscallarg(off_t) offset;
	} */ *uap = v;
	int error, cmd = SCARG(uap, cmd);
	size_t len;
	off_t i, offset;
	ino_t inode = SCARG(uap, inode);
	dev_t device = SCARG(uap, device);
	char path[MAXPATHLEN], md5[MD5_SUM_LEN];
	struct k5_info *ptr;

	/*
	 * Don't let non-privileged users mess with the database.
	 */
	if (suser(p->p_ucred, &p->p_acflag)) return (EPERM);

	switch (cmd) {
	case K5_ADD:
		/*
		 * Copy the path from userspace.
		 */
		bzero(path, MAXPATHLEN);
		error = copyinstr(SCARG(uap, path), path, MAXPATHLEN, &len);
		if (error || !len) {
			uprintf("Stephanie: k5: Invalid path.\n");
			return (error ? error : EINVAL);
		}

		/*
		 * Copy the MD5 checksum from userspace.
		 */
		bzero(md5, MD5_SUM_LEN);
		error = copyinstr(SCARG(uap, md5), md5, MD5_SUM_LEN, &len);
		if (error || len != MD5_SUM_LEN) {
			uprintf("Stephanie: k5: Invalid MD5 checksum.\n");
			return (error ? error : EINVAL);
		}

		if (k5_search(inode, device)) {
			uprintf("Stephanie: k5: Already in database.\n");
			return (EEXIST);
		}

		if ((error = k5_add(inode, device, path, md5))) {
			uprintf("Stephanie: k5: Entry addition "
			    "failed.\n");
			return (error);
		}

#ifdef K5_VERBOSE
		uprintf("Stephanie: k5: Entry added.\n");
#endif

		return (0);

	case K5_DEL:
		if (!(ptr = k5_search(inode, device))) {
			uprintf("Stephanie: k5: No such entry.\n");
			return (ENOENT);
		}

		k5_del(ptr);

#ifdef K5_VERBOSE
		uprintf("Stephanie: k5: Entry removed.\n");
#endif

		return (0);

	case K5_GETINFO:
		if ((ptr = k5_search(inode, device))) {
			error = copyout(ptr, SCARG(uap, ptr),
			    sizeof(struct k5_info));
			if (error) {
				uprintf("Stephanie: k5: Error during data"
				    " copy to userspace.\n");
				return (error);
			}

			return (0);
		}

#ifdef K5_VERBOSE
		uprintf("Stephanie: k5: No such entry.\n");
#endif

		return (ENOENT);

	case K5_GETOFFSET:
		if (SCARG(uap, ptr) == NULL) {
			uprintf("Stephanie: k5: Bad address.\n");
			return (EFAULT);
		}

		offset = SCARG(uap, offset);

		if ((offset < 0) || (offset >= k5_size)) return (EINVAL);

		for (i = 0, ptr = k5_head.lh_first; (ptr != NULL) &&
		    (i < offset); ptr = LIST_NEXT(ptr, k5_list), i++);
		if (ptr) {
			error = copyout(ptr, SCARG(uap, ptr), 
			    sizeof(struct k5_info));
			if (error) {
				uprintf("Stephanie: k5: Error during data"
				    " copy to userspace.\n");
				return (error);
			}

			return (0);
		}

		return (ENOENT);

	default:
		return (EINVAL);
	}

	return (EINVAL);
}
