/*
 * Copyright (c) 2002 br1an <brian@innu.org>
 * All rights reserved.
 *
 * Official Stephanie homepage:
 *
 *	http://innu.org/~brian/Stephanie
 *
 * Shoutz go to shegget, S`, _eci, doe, jaf.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/kern_sneek.h>
#include <sys/vnode.h>

/*
 * Globals: Logging flags and maximum number of arguments to log.
 */
extern unsigned int sneek_flags;
extern unsigned int sneek_maxargs;

/*
 * Initialize execve() logging. Set all flags by default.
 */
void sneek_init(void) {
	sneek_flags = 0;
	sneek_maxargs = 0;

	sneek_flags |=  SNEEK_SET_ON | SNEEK_LOG_ALL;

	printf("Stephanie: Sneek: execve() logging initialized.\n");
}	

/*
 * Print the logging flags set.
 */
void sneek_show(void) {
	uprintf("Stephanie: Sneek: flags=%b", sneek_flags, SNEEK_FBITS);
	
	if ((sneek_flags & SNEEK_LOG_ARGS_MAX) && sneek_maxargs)
		uprintf(" [maxargs=%u]\n", sneek_maxargs);
	else
		uprintf("\n");
}

/*
 * Turn Sneek on or off. This *should* be called before the system goes 
 * multiuser.
 */
void sneek_mode(int mode) {
	extern unsigned int sneek_flags;

	switch (mode) {
	case SNEEK_SET_ON:
		if (sneek_flags & SNEEK_SET_OFF) {
			sneek_flags &= ~SNEEK_SET_OFF;
			uprintf("Stephanie: Sneek: execve() logging is "
			    "now on.\n");
		} else {
			uprintf("Stephanie: Sneek: execve() logging "
			    "already on.\n");
		}

		break;

	case SNEEK_SET_OFF:
		if (!(sneek_flags & SNEEK_SET_OFF)) {
			sneek_flags |= SNEEK_SET_OFF;
			uprintf("Stephanie: Sneek: execve() logging is " 
			    "now off.\n");
		} else {
			uprintf("Stephanie: Sneek: execve() logging "
			    "already off.\n");
		}

		break;
	}
}

/*
 * The logging function. I use a simple concept called ``parent process 
 * traversal'' to make sure I log programs started from within a script in 
 * a noteable fashion, don't log parent process if it's init, etc. to 
 * shave some bytes here and there in the log files..
 */
void sneek_log(struct proc *p, struct exec_package pack,
    char * const *ap, int argc) {
	static int prev_pid;
	int total, curpos, s = splhigh();

	/*
	 * Bump argument pointer to point to first argument.
	 */
	ap++;
	argc--;

	/*
	 * Always works: Log the program that's being executed, and give 
	 * indication if it's a script or not.
	 */
	if ((pack.ep_flags & EXEC_INDIR) &&
	    strncmp(pack.ep_hdr, EXEC_SCRIPT_MAGIC, EXEC_SCRIPT_MAGICLEN)) {
		prev_pid = p->p_pid;
		addlog("(script) \"%s", pack.ep_name);
	} else {
		if ((prev_pid == p->p_pptr->p_pid) ||
		    (prev_pid == p->p_pid))
			addlog(" => \"%s", pack.ep_name);
		else {
			prev_pid = -1;
			addlog("\"%s", pack.ep_name);
		}
	}

	/*
	 * Handle the logging of arguments.
	 */
	if ((sneek_flags & SNEEK_LOG_ARGS) && argc) {
		if ((sneek_flags & SNEEK_LOG_ARGS_MAX) &&
		    (sneek_maxargs < argc)) {
			total = argc;
			argc = sneek_maxargs;
		}

		/*
		 * A ghetto loop to log arguments.
		 */
		for (curpos = 0; curpos < argc; curpos++)
			if (ap[curpos] && ap[curpos][0])
				addlog(" %s", ap[curpos]);
		addlog("\"");

		/*
		 * Provide indication if we logged less arguments than the 
		 * total number of arguments.
		 */
		if(sneek_maxargs && (argc < total))
			addlog(" (%d/%d args logged)", argc, total);
	} else {
		addlog("\"");
	}

	/*
	 * Indication of background processes.
	 */
	if ((sneek_flags & SNEEK_LOG_BG) && !(p->p_flag & P_CONTROLT))
		addlog(" (bg)");

	/*
	 * Take care of logging the entire UID/GID deal. Only log 
	 * effective UID/GID if they are not the same as the 'real' ones.
	 */
	if (sneek_flags & SNEEK_LOG_UIDGID) {
		addlog(" (%u", p->p_cred->p_ruid);
		if (pack.ep_vap->va_mode & VSUID)
			addlog("/%u", pack.ep_vap->va_uid);

		addlog(", %u", p->p_cred->p_rgid);
		if (pack.ep_vap->va_mode & VSGID)
			addlog("/%u", pack.ep_vap->va_gid);
               
		addlog(")");
	}

	/*
	 * Process ID (current + parent) logging.
	 */
	if (sneek_flags & SNEEK_LOG_PID) {
		addlog(" [%d", p->p_pid);
		if (p->p_pptr->p_pid != 1) {
			if (prev_pid == -1)
				addlog(" (%d)", p->p_pptr->p_pid);
			else
				addlog(" (%d)", prev_pid);
		}

		addlog("]");
	}

	/*
	 * Finish with a newline, call logwakeup() to send the data to the 
	 * kernel log device, and restore the system priority level.
	 */
	addlog("\n");
	logwakeup();
	splx(s);
}
